"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.validateExports = void 0;
const models_1 = require("../models");
const ts = require("typescript");
const path_1 = require("path");
function validateExports(project, logger, intentionallyNotExported) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
    const queue = [];
    const context = { reflection: project };
    const seenIntentional = new Set();
    const visitor = makeTypeVisitor(logger, queue, new Set(intentionallyNotExported), seenIntentional, context);
    const add = (item) => {
        if (!item)
            return;
        if (item instanceof models_1.Reflection) {
            queue.push(item);
        }
        else {
            queue.push(...item);
        }
    };
    let current = project;
    do {
        context.reflection = current;
        if (current instanceof models_1.ContainerReflection) {
            add(current.children);
        }
        if (current instanceof models_1.DeclarationReflection) {
            (_a = current.type) === null || _a === void 0 ? void 0 : _a.visit(visitor);
            add(current.typeParameters);
            add(current.signatures);
            add(current.indexSignature);
            add(current.getSignature);
            add(current.setSignature);
            (_b = current.overwrites) === null || _b === void 0 ? void 0 : _b.visit(visitor);
            // Do not check inheritedFrom, it doesn't always make sense to export a base type.
            (_c = current.implementationOf) === null || _c === void 0 ? void 0 : _c.visit(visitor);
            // Do not check extendedTypes, it doesn't always make sense to export a base type.
            // do not validate extendedBy, guaranteed to all be in the documentation.
            (_d = current.implementedTypes) === null || _d === void 0 ? void 0 : _d.forEach((type) => type.visit(visitor));
            // do not validate implementedBy, guaranteed to all be in the documentation.
        }
        if (current instanceof models_1.SignatureReflection) {
            add(current.parameters);
            add(current.typeParameters);
            (_e = current.type) === null || _e === void 0 ? void 0 : _e.visit(visitor);
            (_f = current.overwrites) === null || _f === void 0 ? void 0 : _f.visit(visitor);
            // Do not check inheritedFrom, it doesn't always make sense to export a base type.
            (_g = current.implementationOf) === null || _g === void 0 ? void 0 : _g.visit(visitor);
        }
        if (current instanceof models_1.ParameterReflection) {
            (_h = current.type) === null || _h === void 0 ? void 0 : _h.visit(visitor);
        }
        if (current instanceof models_1.TypeParameterReflection) {
            (_j = current.type) === null || _j === void 0 ? void 0 : _j.visit(visitor);
            (_k = current.default) === null || _k === void 0 ? void 0 : _k.visit(visitor);
        }
    } while ((current = queue.shift()));
    const intentional = new Set(intentionallyNotExported);
    for (const x of seenIntentional) {
        intentional.delete(x);
    }
    if (intentional.size) {
        logger.warn("The following symbols were marked as intentionally not exported, but were either not referenced in the documentation, or were exported:\n\t" +
            Array.from(intentional).join("\n\t"));
    }
}
exports.validateExports = validateExports;
function makeTypeVisitor(logger, queue, intentional, seenIntentional, context) {
    const warned = new Set();
    return (0, models_1.makeRecursiveVisitor)({
        reference(type) {
            var _a, _b;
            // If we don't have a symbol, then this was an intentionally broken reference.
            const symbol = type.getSymbol();
            if (!type.reflection && symbol) {
                if (intentional.has(symbol.name)) {
                    seenIntentional.add(symbol.name);
                }
                if ((symbol.flags & ts.SymbolFlags.TypeParameter) === 0 &&
                    !intentional.has(symbol.name) &&
                    !warned.has(symbol) &&
                    !((_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.some((d) => d.getSourceFile().fileName.includes("node_modules")))) {
                    warned.add(symbol);
                    const decl = (_b = symbol.declarations) === null || _b === void 0 ? void 0 : _b[0];
                    if (decl) {
                        const { line } = ts.getLineAndCharacterOfPosition(decl.getSourceFile(), decl.getStart());
                        const file = (0, path_1.relative)(process.cwd(), decl.getSourceFile().fileName);
                        logger.warn(`${type.name}, defined at ${file}:${line}, is referenced by ${context.reflection.getFullName()} but not included in the documentation.`);
                    }
                }
            }
        },
        reflection(type) {
            queue.push(type.declaration);
        },
    });
}
